# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
import logging
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_machine import CoreSubsets, Router
from pacman.executor.injection_decorator import inject_items
from pacman.model.graphs.common import EdgeTrafficType
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import ConstantSDRAM, ResourceContainer
from spinn_front_end_common.abstract_models import (
    AbstractHasAssociatedBinary, AbstractGeneratesDataSpecification)
from spinn_front_end_common.utilities import globals_variables
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.utility_objs.\
    extra_monitor_scp_processes import (
        ReadStatusProcess, ResetCountersProcess, SetPacketTypesProcess,
        SetRouterTimeoutProcess, ClearQueueProcess,
        LoadApplicationMCRoutesProcess, LoadSystemMCRoutesProcess)
from spinn_front_end_common.utilities.constants import (
    SARK_PER_MALLOC_SDRAM_USAGE, DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES,
    BYTES_PER_WORD, BYTES_PER_KB)
from spinn_front_end_common.utilities.helpful_functions import (
    convert_vertices_to_core_subset)
from spinn_front_end_common.utilities.emergency_recovery import (
    emergency_recover_state_from_failure)
from .data_speed_up_packet_gatherer_machine_vertex import (
    DataSpeedUpPacketGatherMachineVertex as
    Gatherer)

log = FormatAdapter(logging.getLogger(__name__))

_CONFIG_REGION_REINJECTOR_SIZE_IN_BYTES = 5 * BYTES_PER_WORD
#: 1.new seq key, 2.first data key, 3. transaction id key 4.end flag key,
# 5.base key
_CONFIG_DATA_SPEED_UP_SIZE_IN_BYTES = 5 * BYTES_PER_WORD
_CONFIG_MAX_EXTRA_SEQ_NUM_SIZE_IN_BYTES = 460 * BYTES_PER_KB
_CONFIG_DATA_IN_KEYS_SDRAM_IN_BYTES = 3 * BYTES_PER_WORD
_MAX_DATA_SIZE_FOR_DATA_IN_MULTICAST_ROUTING = ((49 * 3) + 1) * BYTES_PER_WORD
_BIT_SHIFT_TO_MOVE_APP_ID = 24

# cap for stopping wrap arounds
TRANSACTION_ID_CAP = 0xFFFFFFFF

# SDRAM requirement for containing router table entries
# 16 bytes per entry:
# 4 for a key, 4 for mask,
# 4 for word alignment for 18 cores and 6 links
# (24 bits, for word aligning)
_SDRAM_FOR_ROUTER_TABLE_ENTRIES = 1024 * 4 * BYTES_PER_WORD


class _DSG_REGIONS(Enum):
    REINJECT_CONFIG = 0
    DATA_OUT_CONFIG = 1
    DATA_IN_CONFIG = 2


class _KEY_OFFSETS(Enum):
    ADDRESS_KEY_OFFSET = 0
    DATA_KEY_OFFSET = 1
    BOUNDARY_KEY_OFFSET = 2


class ExtraMonitorSupportMachineVertex(
        MachineVertex, AbstractHasAssociatedBinary,
        AbstractGeneratesDataSpecification):
    """ Machine vertex for talking to extra monitor cores. \
        Supports reinjection control and the faster data transfer protocols.

    Usually deployed once per chip.
    """

    __slots__ = (
        # if we reinject multicast packets
        "_reinject_multicast",
        # if we reinject point to point packets
        "_reinject_point_to_point",
        # if we reinject nearest neighbour packets
        "_reinject_nearest_neighbour",
        # if we reinject fixed route packets
        "_reinject_fixed_route",
        # placement holder for ease of access
        "_placement",
        # app id, used for reporting failures on system core RTE
        "_app_id",
        # machine instance
        "_machine",
        # the local transaction id
        "_transaction_id"
    )

    def __init__(
            self, constraints, app_vertex,
            reinject_multicast=None, reinject_point_to_point=False,
            reinject_nearest_neighbour=False, reinject_fixed_route=False):
        """
        :param constraints: constraints on this vertex
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint)
        :param bool reinject_multicast:
            if we reinject multicast packets; defaults to value of
            `enable_reinjection` setting in configuration file
        :param bool reinject_point_to_point:
            if we reinject point-to-point packets
        :param bool reinject_nearest_neighbour:
            if we reinject nearest-neighbour packets
        :param bool reinject_fixed_route: if we reinject fixed route packets
        """
        # pylint: disable=too-many-arguments
        super().__init__(
            label="SYSTEM:ExtraMonitor", constraints=constraints,
            app_vertex=app_vertex)

        if reinject_multicast is None:
            config = globals_variables.get_simulator().config
            self._reinject_multicast = config.getboolean(
                "Machine", "enable_reinjection")
        else:
            self._reinject_multicast = reinject_multicast
        self._reinject_point_to_point = reinject_point_to_point
        self._reinject_nearest_neighbour = reinject_nearest_neighbour
        self._reinject_fixed_route = reinject_fixed_route
        # placement holder for ease of access
        self._placement = None
        self._app_id = None
        self._machine = None
        self._transaction_id = 0

    @property
    def reinject_multicast(self):
        """
        :rtype: bool
        """
        return self._reinject_multicast

    @property
    def transaction_id(self):
        return self._transaction_id

    def update_transaction_id(self):
        self._transaction_id = (self._transaction_id + 1) & TRANSACTION_ID_CAP

    def update_transaction_id_from_machine(self, txrx):
        """ looks up from the machine what the current transaction id is
        and updates the extra monitor.

        :param txrx: SpiNNMan instance
        :rtype: None
        """
        self._transaction_id = txrx.read_user_1(
            self._placement.x, self._placement.y, self._placement.p)

    @property
    def reinject_point_to_point(self):
        """
        :rtype: bool
        """
        return self._reinject_point_to_point

    @property
    def reinject_nearest_neighbour(self):
        """
        :rtype: bool
        """
        return self._reinject_nearest_neighbour

    @property
    def reinject_fixed_route(self):
        """
        :rtype: bool
        """
        return self._reinject_fixed_route

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        return self.static_resources_required()

    @property
    def placement(self):
        """
        :rtype: ~pacman.model.placements.Placement
        """
        return self._placement

    @staticmethod
    def static_resources_required():
        """ The resources required by this vertex.

        :rtype: ~pacman.model.resources.ResourceContainer
        """
        return ResourceContainer(sdram=ConstantSDRAM(
            _CONFIG_REGION_REINJECTOR_SIZE_IN_BYTES +
            _CONFIG_DATA_SPEED_UP_SIZE_IN_BYTES +
            _CONFIG_MAX_EXTRA_SEQ_NUM_SIZE_IN_BYTES +
            # Data spec size
            DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES +
            # One malloc for extra sequence numbers
            SARK_PER_MALLOC_SDRAM_USAGE +
            _MAX_DATA_SIZE_FOR_DATA_IN_MULTICAST_ROUTING +
            _SDRAM_FOR_ROUTER_TABLE_ENTRIES +
            _CONFIG_DATA_IN_KEYS_SDRAM_IN_BYTES))

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return self.static_get_binary_start_type()

    @staticmethod
    def static_get_binary_start_type():
        """ The type of the binary implementing this vertex.

        :rtype: ExecutableType
        """
        return ExecutableType.SYSTEM

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return self.static_get_binary_file_name()

    @staticmethod
    def static_get_binary_file_name():
        """ The name of the binary implementing this vertex.

        :rtype: str
        """
        return "extra_monitor_support.aplx"

    @inject_items({"routing_info": "MemoryRoutingInfos",
                   "machine_graph": "MemoryMachineGraph",
                   "data_in_routing_tables": "DataInMulticastRoutingTables",
                   "mc_data_chips_to_keys": "DataInMulticastKeyToChipMap",
                   "app_id": "APPID",
                   "machine": "MemoryExtendedMachine",
                   "router_timeout_keys": "SystemMulticastRouterTimeoutKeys"})
    @overrides(AbstractGeneratesDataSpecification.generate_data_specification,
               additional_arguments={
                   "routing_info", "machine_graph", "data_in_routing_tables",
                   "mc_data_chips_to_keys", "app_id", "machine",
                   "router_timeout_keys"})
    def generate_data_specification(
            self, spec, placement, routing_info, machine_graph,
            data_in_routing_tables, mc_data_chips_to_keys, app_id,
            machine, router_timeout_keys):
        """
        :param ~pacman.model.routing_info.RoutingInfo routing_info: (injected)
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            (injected)
        :param data_in_routing_tables: (injected)
        :type data_in_routing_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param dict(tuple(int,int),int) mc_data_chips_to_keys: (injected)
        :param int app_id: (injected)
        :param ~spinn_machine.Machine machine: (injected)
        """
        # pylint: disable=arguments-differ
        # storing for future usage
        self._placement = placement
        self._app_id = app_id
        self._machine = machine
        # write reinjection config
        self._generate_reinjection_config(
            spec, router_timeout_keys, placement, machine)
        # write data speed up out config
        self._generate_data_speed_up_out_config(
            spec, routing_info, machine_graph)
        # write data speed up in config
        self._generate_data_speed_up_in_config(
            spec, data_in_routing_tables,
            machine.get_chip_at(placement.x, placement.y),
            mc_data_chips_to_keys)
        spec.end_specification()

    def _generate_data_speed_up_out_config(
            self, spec, routing_info, machine_graph):
        """
        :param ~.DataSpecificationGenerator spec: spec file
        :param ~.RoutingInfo routing_info: the packet routing info
        :param ~.MachineGraph machine_graph: The graph containing this vertex
        """
        spec.reserve_memory_region(
            region=_DSG_REGIONS.DATA_OUT_CONFIG.value,
            size=_CONFIG_DATA_SPEED_UP_SIZE_IN_BYTES,
            label="data speed-up out config region")
        spec.switch_write_focus(_DSG_REGIONS.DATA_OUT_CONFIG.value)

        if Gatherer.TRAFFIC_TYPE == EdgeTrafficType.MULTICAST:
            base_key = routing_info.get_first_key_for_edge(
                list(machine_graph.get_edges_starting_at_vertex(self))[0])
            spec.write_value(base_key)
            spec.write_value(base_key + Gatherer.NEW_SEQ_KEY_OFFSET)
            spec.write_value(base_key + Gatherer.FIRST_DATA_KEY_OFFSET)
            spec.write_value(base_key + Gatherer.TRANSACTION_ID_KEY_OFFSET)
            spec.write_value(base_key + Gatherer.END_FLAG_KEY_OFFSET)
        else:
            spec.write_value(Gatherer.BASE_KEY)
            spec.write_value(Gatherer.NEW_SEQ_KEY)
            spec.write_value(Gatherer.FIRST_DATA_KEY)
            spec.write_value(Gatherer.TRANSACTION_ID_KEY)
            spec.write_value(Gatherer.END_FLAG_KEY)

    def _generate_reinjection_config(
            self, spec, router_timeout_keys, placement, machine):
        """
        :param ~.DataSpecificationGenerator spec: spec file
        :param dict(tuple(int,int),int) router_timeout_keys:
        :param ~.Placement placement:
        :param ~.Machine machine:
        """
        spec.reserve_memory_region(
            region=_DSG_REGIONS.REINJECT_CONFIG.value,
            size=_CONFIG_REGION_REINJECTOR_SIZE_IN_BYTES,
            label="re-injection config region")

        spec.switch_write_focus(_DSG_REGIONS.REINJECT_CONFIG.value)
        for value in [
                self._reinject_multicast, self._reinject_point_to_point,
                self._reinject_fixed_route,
                self._reinject_nearest_neighbour]:
            # Note that this is inverted! Why... I dunno!
            spec.write_value(int(not value))

        # add the reinjection mc interface
        chip = machine.get_chip_at(placement.x, placement.y)
        reinjector_base_mc_key = (
            router_timeout_keys[
                (chip.nearest_ethernet_x, chip.nearest_ethernet_y)])
        spec.write_value(reinjector_base_mc_key)

    def _generate_data_speed_up_in_config(
            self, spec, data_in_routing_tables, chip, mc_data_chips_to_keys):
        """
        :param ~.DataSpecificationGenerator spec: spec file
        :param ~.MulticastRoutingTables data_in_routing_tables:
            routing tables for all chips
        :param ~.Chip chip: the chip where this monitor will run
        :param dict(tuple(int,int),int) mc_data_chips_to_keys:
            data in keys to chips map.
        """
        spec.reserve_memory_region(
            region=_DSG_REGIONS.DATA_IN_CONFIG.value,
            size=(_MAX_DATA_SIZE_FOR_DATA_IN_MULTICAST_ROUTING +
                  _CONFIG_DATA_IN_KEYS_SDRAM_IN_BYTES),
            label="data speed-up in config region")
        spec.switch_write_focus(_DSG_REGIONS.DATA_IN_CONFIG.value)

        # write address key and data key
        base_key = mc_data_chips_to_keys[chip.x, chip.y]
        spec.write_value(base_key + _KEY_OFFSETS.ADDRESS_KEY_OFFSET.value)
        spec.write_value(base_key + _KEY_OFFSETS.DATA_KEY_OFFSET.value)
        spec.write_value(base_key + _KEY_OFFSETS.BOUNDARY_KEY_OFFSET.value)

        # write table entries
        table = data_in_routing_tables.get_routing_table_for_chip(
            chip.x, chip.y)
        spec.write_value(table.number_of_entries)
        for entry in table.multicast_routing_entries:
            spec.write_value(entry.routing_entry_key)
            spec.write_value(entry.mask)
            spec.write_value(self.__encode_route(entry))

    def __encode_route(self, entry):
        """
        :param ~spinn_machine.MulticastRoutingEntry entry:
        :rtype: int
        """
        route = self._app_id << _BIT_SHIFT_TO_MOVE_APP_ID
        route |= Router.convert_routing_table_entry_to_spinnaker_route(entry)
        return route

    def set_router_wait1_timeout(
            self, timeout, transceiver, placements,
            extra_monitor_cores_to_set):
        """ Supports setting of the router time outs for a set of chips via\
            their extra monitor cores. This sets the timeout for the time\
            between when a packet arrives and when it starts to be emergency\
            routed. (Actual emergency routing is disabled by default.)

        :param tuple(int,int) timeout:
            The mantissa and exponent of the timeout value, each between
            0 and 15
        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnman interface
        :param ~pacman.model.placements.Placements placements:
            vertex placements
        :param extra_monitor_cores_to_set:
            which monitors control the routers to set the timeout of
        :type extra_monitor_cores_to_set:
            iterable(ExtraMonitorSupportMachineVertex)
        """
        mantissa, exponent = timeout
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_to_set, placements)
        process = SetRouterTimeoutProcess(
            transceiver.scamp_connection_selector)
        try:
            process.set_wait1_timeout(mantissa, exponent, core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def set_router_wait2_timeout(
            self, timeout, transceiver, placements,
            extra_monitor_cores_to_set):
        """ Supports setting of the router time outs for a set of chips via\
            their extra monitor cores. This sets the timeout for the time\
            between when a packet starts to be emergency routed and when it\
            is dropped. (Actual emergency routing is disabled by default.)

        :param tuple(int,int) timeout:
            The mantissa and exponent of the timeout value, each between
            0 and 15
        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan instance
        :param ~pacman.model.placements.Placements placements:
            vertex placements
        :param extra_monitor_cores_to_set:
            which monitors control the routers to set the timeout of
        :type extra_monitor_cores_to_set:
            iterable(ExtraMonitorSupportMachineVertex)
        """
        mantissa, exponent = timeout
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_to_set, placements)
        process = SetRouterTimeoutProcess(
            transceiver.scamp_connection_selector)
        try:
            process.set_wait2_timeout(mantissa, exponent, core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def reset_reinjection_counters(
            self, transceiver, placements, extra_monitor_cores_to_set):
        """ Resets the counters for reinjection

        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        :param ~pacman.model.placements.Placements placements:
            the placements object
        :param extra_monitor_cores_to_set:
            which monitors control the routers to reset the counters of
        :type extra_monitor_cores_to_set:
            iterable(ExtraMonitorSupportMachineVertex)
        """
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_to_set, placements)
        process = ResetCountersProcess(transceiver.scamp_connection_selector)
        try:
            process.reset_counters(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def clear_reinjection_queue(
            self, transceiver, placements, extra_monitor_cores_to_set):
        """ Clears the queues for reinjection

        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        :param ~pacman.model.placements.Placements placements:
            the placements object
        :param extra_monitor_cores_to_set:
            Which extra monitors need to clear their queues.
        :type extra_monitor_cores_to_set:
            iterable(ExtraMonitorSupportMachineVertex)
        """
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_to_set, placements)
        process = ClearQueueProcess(transceiver.scamp_connection_selector)
        try:
            process.reset_counters(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def get_reinjection_status(self, placements, transceiver):
        """ Get the reinjection status from this extra monitor vertex

        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        :param ~pacman.model.placements.Placements placements:
            the vertex placements
        :return: the reinjection status for this vertex
        :rtype: ReInjectionStatus
        """
        placement = placements.get_placement_of_vertex(self)
        process = ReadStatusProcess(transceiver.scamp_connection_selector)
        try:
            return process.get_reinjection_status(
                placement.x, placement.y, placement.p)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self, placement)
            raise

    def get_reinjection_status_for_vertices(
            self, placements, extra_monitor_cores_for_data, transceiver):
        """ Get the reinjection status from a set of extra monitor cores

        :param ~pacman.model.placements.Placements placements:
            the placements object
        :param extra_monitor_cores_for_data:
            the extra monitor cores to get status from
        :type extra_monitor_cores_for_data:
            iterable(ExtraMonitorSupportMachineVertex)
        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        :rtype: dict(tuple(int,int), ReInjectionStatus)
        """
        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_for_data, placements)
        process = ReadStatusProcess(transceiver.scamp_connection_selector)
        return process.get_reinjection_status_for_core_subsets(core_subsets)

    def set_reinjection_packets(
            self, placements, extra_monitor_cores_for_data, transceiver,
            point_to_point=None, multicast=None, nearest_neighbour=None,
            fixed_route=None):
        """
        :param ~pacman.model.placements.Placements placements:
            placements object
        :param extra_monitor_cores_for_data:
            the extra monitor cores to set the packets of
        :type extra_monitor_cores_for_data:
            iterable(ExtraMonitorSupportMachineVertex)
        :param ~spinnman.transceiver.Transceiver transceiver: spinnman instance
        :param point_to_point:
            If point to point should be set, or None if left as before
        :type point_to_point: bool or None
        :param multicast:
            If multicast should be set, or None if left as before
        :type multicast: bool or None
        :param nearest_neighbour:
            If nearest neighbour should be set, or None if left as before
        :type nearest_neighbour: bool or None
        :param fixed_route:
            If fixed route should be set, or None if left as before.
        :type fixed_route: bool or None
        """
        # pylint: disable=too-many-arguments
        if multicast is not None:
            self._reinject_multicast = multicast
        if point_to_point is not None:
            self._reinject_point_to_point = point_to_point
        if nearest_neighbour is not None:
            self._reinject_nearest_neighbour = nearest_neighbour
        if fixed_route is not None:
            self._reinject_fixed_route = fixed_route

        core_subsets = convert_vertices_to_core_subset(
            extra_monitor_cores_for_data, placements)
        process = SetPacketTypesProcess(transceiver.scamp_connection_selector)
        try:
            process.set_packet_types(
                core_subsets, self._reinject_point_to_point,
                self._reinject_multicast, self._reinject_nearest_neighbour,
                self._reinject_fixed_route)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def load_system_mc_routes(
            self, placements, extra_monitor_cores_for_data, transceiver):
        """ Get the extra monitor cores to load up the system-based \
            multicast routes (used by data in).

        :param ~pacman.model.placements.Placements placements:
            the placements object
        :param extra_monitor_cores_for_data:
            the extra monitor cores to get status from
        :type extra_monitor_cores_for_data:
            iterable(ExtraMonitorSupportMachineVertex)
        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        """
        core_subsets = self._convert_vertices_to_core_subset(
            extra_monitor_cores_for_data, placements)
        process = LoadSystemMCRoutesProcess(
            transceiver.scamp_connection_selector)
        try:
            return process.load_system_mc_routes(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def load_application_mc_routes(
            self, placements, extra_monitor_cores_for_data, transceiver):
        """ Get the extra monitor cores to load up the application-based\
            multicast routes (used by data in).

        :param ~pacman.model.placements.Placements placements:
            the placements object
        :param extra_monitor_cores_for_data:
            the extra monitor cores to get status from
        :type extra_monitor_cores_for_data:
            iterable(ExtraMonitorSupportMachineVertex)
        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        """
        core_subsets = self._convert_vertices_to_core_subset(
            extra_monitor_cores_for_data, placements)
        process = LoadApplicationMCRoutesProcess(
            transceiver.scamp_connection_selector)
        try:
            return process.load_application_mc_routes(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    @staticmethod
    def _convert_vertices_to_core_subset(extra_monitor_cores, placements):
        """ Convert vertices into the subset of cores where they've been\
            placed.

        :param iterable(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
            the vertices to convert to core subsets
        :param ~.Placements placements: the placements object
        :return: where the vertices have been placed
        :rtype: ~.CoreSubsets
        """
        core_subsets = CoreSubsets()
        for vertex in extra_monitor_cores:
            if not isinstance(vertex, ExtraMonitorSupportMachineVertex):
                raise Exception(
                    "can only use ExtraMonitorSupportMachineVertex to set "
                    "the router time out")
            placement = placements.get_placement_of_vertex(vertex)
            core_subsets.add_processor(placement.x, placement.y, placement.p)
        return core_subsets



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from spinn_utilities.log import FormatAdapter
from spinnman.connections import SocketAddressWithChip
from spinnman.constants import POWER_CYCLE_WAIT_TIME_IN_SECONDS
from spinnman.transceiver import create_transceiver_from_hostname
from spinnman.model import BMPConnectionData
from spinn_front_end_common.utilities.exceptions import ConfigurationException
import time
import logging

logger = FormatAdapter(logging.getLogger(__name__))


class MachineGenerator(object):
    """ Makes a transceiver and a machine object.
    """

    POWER_CYCLE_WARNING = (
        "When power-cycling a board, it is recommended that you wait for 30 "
        "seconds before attempting a reboot. Therefore, the tools will now "
        "wait for 30 seconds. If you wish to avoid this wait, please set "
        "reset_machine_on_startup = False in the [Machine] section of the "
        "relevant configuration (cfg) file.")

    POWER_CYCLE_FAILURE_WARNING = (
        "The end user requested the power-cycling of the board. But the "
        "tools did not have the required BMP connection to facilitate a "
        "power-cycling, and therefore will not do so. please set the "
        "bmp_names accordingly in the [Machine] section of the relevant "
        "configuration (cfg) file. Or use a machine assess process which "
        "provides the BMP data (such as a spalloc system) or finally set "
        "reset_machine_on_startup = False in the [Machine] section of the "
        "relevant configuration (cfg) file to avoid this warning in future.")

    __slots__ = []

    def __call__(
            self, hostname, bmp_details, downed_chips, downed_cores,
            downed_links, board_version, auto_detect_bmp,
            scamp_connection_data, boot_port_num, reset_machine_on_start_up,
            report_waiting_logs, max_sdram_size=None, repair_machine=False,
            ignore_bad_ethernets=True, default_report_directory=None):
        """
        :param str hostname:
            the hostname or IP address of the SpiNNaker machine
        :param str bmp_details: the details of the BMP connections
        :param set(tuple(int,int)) downed_chips:
            the chips that are down which SARK thinks are alive
        :param set(tuple(int,int,int)) downed_cores:
            the cores that are down which SARK thinks are alive
        :param set(tuple(int,int,int)) downed_links:
            the links that are down which SARK thinks are alive
        :param int board_version:
            the version of the boards being used within the machine
            (1, 2, 3, 4 or 5)
        :param bool auto_detect_bmp:
            Whether the BMP should be automatically determined
        :param scamp_connection_data:
            the list of SC&MP connection data or None
        :type scamp_connection_data:
            list(~spinnman.connections.SocketAddressWithChip)
        :param int boot_port_num: the port number used for the boot connection
        :param bool reset_machine_on_start_up:
        :param max_sdram_size: the maximum SDRAM each chip can say it has
            (mainly used in debugging purposes)
        :type max_sdram_size: int or None
        :param bool repair_machine:
            Flag to set the behaviour if a repairable error is found on the
            machine.
            If `True` will create a machine without the problematic bits.
            (See machine_factory.machine_repair)
            If `False`, get machine will raise an Exception if a problematic
            machine is discovered.
        :param bool ignore_bad_ethernets:
            Flag to say that ip_address information
            on non-ethernet chips should be ignored.
            Non-ethernet chips are defined here as ones that do not report
            themselves their nearest ethernet.
            The bad IP address is always logged.
            If True, the IP address is ignored.
            If False, the chip with the bad IP address is removed.
        :param str default_report_directory:
            Directory to write any reports too.
            If None the current directory will be used.
        :param report_waiting_logs:
            flag for the txrx to report when waiting on busy cores.
        :type report_waiting_logs: bool
        :return: Transceiver, and description of machine it is connected to
        :rtype: tuple(~spinnman.transceiver.Transceiver,
            ~spinn_machine.Machine)
        """
        # pylint: disable=too-many-arguments

        # if the end user gives you SCAMP data, use it and don't discover them
        if scamp_connection_data is not None:
            scamp_connection_data = [
                self._parse_scamp_connection(piece)
                for piece in scamp_connection_data.split(":")]

        txrx = create_transceiver_from_hostname(
            hostname=hostname,
            bmp_connection_data=self._parse_bmp_details(bmp_details),
            version=board_version, ignore_chips=downed_chips,
            ignore_cores=downed_cores, ignored_links=downed_links,
            auto_detect_bmp=auto_detect_bmp, boot_port_no=boot_port_num,
            scamp_connections=scamp_connection_data,
            max_sdram_size=max_sdram_size,
            repair_machine=repair_machine,
            ignore_bad_ethernets=ignore_bad_ethernets,
            default_report_directory=default_report_directory,
            report_waiting_logs=report_waiting_logs)

        if reset_machine_on_start_up:
            success = txrx.power_off_machine()
            if success:
                logger.warning(self.POWER_CYCLE_WARNING)
                time.sleep(POWER_CYCLE_WAIT_TIME_IN_SECONDS)
                logger.warning("Power cycle wait complete")
            else:
                logger.warning(self.POWER_CYCLE_FAILURE_WARNING)

        # do auto boot if possible
        if board_version is None:
            raise ConfigurationException(
                "Please set a machine version number in the "
                "corresponding configuration (cfg) file")
        txrx.ensure_board_is_ready()
        txrx.discover_scamp_connections()
        return txrx.get_machine_details(), txrx

    @staticmethod
    def _parse_scamp_connection(scamp_connection):
        """
        :param str scamp_connection:
        :rtype: ~.SocketAddressWithChip
        :raises Exception: If the parse fails
        """
        pieces = scamp_connection.split(",")
        if len(pieces) == 3:
            port_num = None
            hostname, chip_x, chip_y = pieces
        elif len(pieces) == 4:
            hostname, port_num, chip_x, chip_y = pieces
        else:
            raise Exception("bad SC&MP connection descriptor")

        return SocketAddressWithChip(
            hostname=hostname,
            port_num=None if port_num is None else int(port_num),
            chip_x=int(chip_x),
            chip_y=int(chip_y))

    @staticmethod
    def _parse_bmp_cabinet_and_frame(bmp_cabinet_and_frame):
        """
        :param str bmp_cabinet_and_frame:
        :rtype: tuple(int or str, int or str, str, str or None)
        """
        split_string = bmp_cabinet_and_frame.split(";", 2)
        if len(split_string) == 1:
            host = split_string[0].split(",")
            if len(host) == 1:
                return 0, 0, split_string[0], None
            return 0, 0, host[0], host[1]
        if len(split_string) == 2:
            host = split_string[1].split(",")
            if len(host) == 1:
                return 0, split_string[0], host[0], None
            return 0, split_string[0], host[0], host[1]
        host = split_string[2].split(",")
        if len(host) == 1:
            return split_string[0], split_string[1], host[0], None
        return split_string[0], split_string[1], host[0], host[1]

    @staticmethod
    def _parse_bmp_boards(bmp_boards):
        """
        :param str bmp_boards:
        :rtype: list(int)
        """
        # If the string is a range of boards, get the range
        range_match = re.match(r"(\d+)-(\d+)", bmp_boards)
        if range_match is not None:
            return list(range(int(range_match.group(1)),
                              int(range_match.group(2)) + 1))

        # Otherwise, assume a list of boards
        return [int(board) for board in bmp_boards.split(",")]

    def _parse_bmp_connection(self, bmp_detail):
        """ Parses one item of BMP connection data. Maximal format:\
            `cabinet;frame;host,port/boards`

        All parts except host can be omitted. Boards can be a \
        hyphen-separated range or a comma-separated list.

        :param str bmp_detail:
        :rtype: ~.BMPConnectionData
        """
        pieces = bmp_detail.split("/")
        (cabinet, frame, hostname, port_num) = \
            self._parse_bmp_cabinet_and_frame(pieces[0])
        # if there is no split, then assume its one board, located at 0
        boards = [0] if len(pieces) == 1 else self._parse_bmp_boards(pieces[1])
        port_num = None if port_num is None else int(port_num)
        return BMPConnectionData(cabinet, frame, hostname, boards, port_num)

    def _parse_bmp_details(self, bmp_string):
        """ Take a BMP line (a colon-separated list) and split it into the\
            BMP connection data.

        :param str bmp_string: the BMP string to be converted
        :return: the BMP connection data
        :rtype: list(~.BMPConnectionData) or None
        """
        if bmp_string is None or bmp_string == "None":
            return None
        return [self._parse_bmp_connection(bmp_connection)
                for bmp_connection in bmp_string.split(":")]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import abstractmethod
from spinn_utilities.overrides import overrides
from pacman.executor.injection_decorator import (
    supports_injection, inject_items)
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification)


@supports_injection
class MachineDataSpecableVertex(
        AbstractGeneratesDataSpecification, allow_derivation=True):
    """ Support for a vertex that simplifies generating a data specification.
    """
    __slots__ = ()

    @inject_items({
        "machine_graph": "MemoryMachineGraph",
        "routing_info": "MemoryRoutingInfos",
        "tags": "MemoryTags",
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor"
    })
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_graph", "routing_info", "tags",
            "machine_time_step", "time_scale_factor"
        })
    def generate_data_specification(
            self, spec, placement, machine_graph, routing_info, tags,
            machine_time_step, time_scale_factor):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            (Injected)
        :param ~pacman.model.routing_info.RoutingInfo routing_info: (Injected)
        :param ~pacman.model.tags.Tags tags: (Injected)
        :param int machine_time_step: (Injected)
        :param int time_scale_factor: (Injected)
        """
        # pylint: disable=too-many-arguments, arguments-differ
        iptags = tags.get_ip_tags_for_vertex(placement.vertex)
        reverse_iptags = tags.get_reverse_ip_tags_for_vertex(placement.vertex)
        self.generate_machine_data_specification(
            spec, placement, machine_graph, routing_info, iptags,
            reverse_iptags, machine_time_step, time_scale_factor)

    @abstractmethod
    def generate_machine_data_specification(
            self, spec, placement, machine_graph, routing_info, iptags,
            reverse_iptags, machine_time_step, time_scale_factor):
        """
        :param ~data_specification.DataSpecificationGenerator spec:
            The data specification to write into.
        :param ~pacman.model.placements.Placement placement:
            Where this node is on the SpiNNaker machine.
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            The graph containing this node.
        :param ~pacman.model.routing_info.RoutingInfo routing_info:
            The routing info.
        :param iptags: The (forward) IP tags for the vertex, if any
        :type iptags: iterable(~spinn_machine.tags.IPTag) or None
        :param reverse_iptags: The reverse IP tags for the vertex, if any
        :type reverse_iptags:
            iterable(~spinn_machine.tags.ReverseIPTag) or None
        :param int machine_time_step: The machine time step
        :param int time_scale_factor: The time step scaling factor
        :rtype: None
        """
        # pylint: disable=too-many-arguments



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.constraints.placer_constraints import ChipAndCoreConstraint
from spinn_front_end_common.utility_models import (
    DataSpeedUpPacketGather, DataSpeedUpPacketGatherMachineVertex,
    ExtraMonitorSupport, ExtraMonitorSupportMachineVertex)


class InsertExtraMonitorVerticesToGraphs(object):
    """ Inserts the extra monitor vertices into the graph that correspond to\
        the extra monitor cores required.
    """

    __slots__ = [
        # the directory where reports go
        "_report_dir",
        #  flag indicating whether to write the reports for data speed up
        "_write_reports",
    ]

    def __call__(
            self, machine, machine_graph, default_report_directory,
            write_data_speed_up_reports, application_graph=None):
        """
        :param ~spinn_machine.Machine machine: spinnMachine instance
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            machine graph
        :param str default_report_directory: the directory where reports go
        :param bool write_data_speed_up_reports:
            determine whether to write the reports for data speed up
        :param int n_cores_to_allocate:
            number of cores to allocate for reception
        :param application_graph: app graph
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        :return: vertex to Ethernet connection map,
            list of extra_monitor_vertices,
            vertex_to_chip_map
        :rtype: tuple(
            dict(tuple(int,int),DataSpeedUpPacketGatherMachineVertex),
            list(ExtraMonitorSupportMachineVertex),
            dict(tuple(int,int),ExtraMonitorSupportMachineVertex))
        """
        # pylint: disable=too-many-arguments, attribute-defined-outside-init
        self._report_dir = default_report_directory
        self._write_reports = write_data_speed_up_reports

        progress = ProgressBar(
            machine.n_chips + len(list(machine.ethernet_connected_chips)),
            "Inserting extra monitors into graphs")

        chip_to_gatherer_map = dict()
        vertex_to_chip_map = dict()

        # handle reinjector and chip based data extractor functionality.
        if application_graph is not None:
            extra_monitors = self._add_second_monitors_application_graph(
                progress, machine, application_graph, machine_graph,
                vertex_to_chip_map)
        else:
            extra_monitors = self._add_second_monitors_machine_graph(
                progress, machine, machine_graph, vertex_to_chip_map)

        # progress data receiver for data extraction functionality
        if application_graph is not None:
            self._add_data_extraction_vertices_app_graph(
                progress, machine, application_graph, machine_graph,
                chip_to_gatherer_map, vertex_to_chip_map)
        else:
            self._add_data_extraction_vertices_mach_graph(
                progress, machine, machine_graph, chip_to_gatherer_map,
                vertex_to_chip_map)

        return chip_to_gatherer_map, extra_monitors, vertex_to_chip_map

    def _add_second_monitors_application_graph(
            self, progress, machine, application_graph, machine_graph,
            vertex_to_chip_map):
        """ Handles placing the second monitor vertex with extra functionality\
            into the graph

        :param ~.ProgressBar progress: progress bar
        :param ~.Machine machine: spinnMachine instance
        :param ~.ApplicationGraph application_graph: app graph
        :param ~.MachineGraph machine_graph: machine graph
        :param dict vertex_to_chip_map: map between vertex and chip
        :return: list of extra monitor vertices
        :rtype: list(~.MachineVertex)
        """
        # pylint: disable=too-many-arguments

        extra_monitor_vertices = list()

        for chip in progress.over(machine.chips, finish_at_end=False):
            if chip.virtual:
                continue
            # add to both application graph and machine graph
            app_vertex = self.__new_app_monitor(chip)
            application_graph.add_vertex(app_vertex)
            machine_vertex = app_vertex.machine_vertex
            machine_graph.add_vertex(machine_vertex)
            vertex_to_chip_map[chip.x, chip.y] = machine_vertex
            extra_monitor_vertices.append(machine_vertex)

        return extra_monitor_vertices

    def _add_second_monitors_machine_graph(
            self, progress, machine, machine_graph, vertex_to_chip_map):
        """ Handles placing the second monitor vertex with extra functionality\
            into the graph

        :param ~.ProgressBar progress: progress bar
        :param ~.Machine machine: spinnMachine instance
        :param ~.MachineGraph machine_graph: machine graph
        :param dict vertex_to_chip_map: map between vertex and chip
        :return: list of extra monitor vertices
        :rtype: list(~.MachineVertex)
        """
        # pylint: disable=too-many-arguments

        extra_monitor_vertices = list()

        for chip in progress.over(machine.chips, finish_at_end=False):
            if chip.virtual:
                continue
            # add to machine graph
            vertex = self.__new_mach_monitor(chip)
            machine_graph.add_vertex(vertex)
            vertex_to_chip_map[chip.x, chip.y] = vertex
            extra_monitor_vertices.append(vertex)

        return extra_monitor_vertices

    def _add_data_extraction_vertices_app_graph(
            self, progress, machine, application_graph, machine_graph,
            chip_to_gatherer_map, vertex_to_chip_map):
        """ Places vertices for receiving data extraction packets.

        :param ~.ProgressBar progress: progress bar
        :param ~.Machine machine: machine instance
        :param ~.ApplicationGraph application_graph: application graph
        :param ~.MachineGraph machine_graph: machine graph
        :param dict chip_to_gatherer_map: vertex to chip map
        :param dict vertex_to_chip_map: map between chip and extra monitor
        """
        # pylint: disable=too-many-arguments

        # insert machine vertices
        for chip in progress.over(machine.ethernet_connected_chips):
            # add to application graph
            app_vertex = self.__new_app_gatherer(chip, vertex_to_chip_map)
            application_graph.add_vertex(app_vertex)
            machine_vertex = app_vertex.machine_vertex
            machine_graph.add_vertex(machine_vertex)
            # update mapping for edge builder
            chip_to_gatherer_map[chip.x, chip.y] = machine_vertex

    def _add_data_extraction_vertices_mach_graph(
            self, progress, machine, machine_graph,
            chip_to_gatherer_map, vertex_to_chip_map):
        """ Places vertices for receiving data extraction packets.

        :param ~.ProgressBar progress: progress bar
        :param ~.Machine machine: machine instance
        :param ~.MachineGraph machine_graph: machine graph
        :param dict chip_to_gatherer_map: vertex to chip map
        :param dict vertex_to_chip_map: map between chip and extra monitor
        """
        # pylint: disable=too-many-arguments

        # insert machine vertices
        for chip in progress.over(machine.ethernet_connected_chips):
            machine_vertex = self.__new_mach_gatherer(chip, vertex_to_chip_map)
            machine_graph.add_vertex(machine_vertex)
            # update mapping for edge builder
            chip_to_gatherer_map[chip.x, chip.y] = machine_vertex

    @staticmethod
    def __new_app_monitor(chip):
        """
        :param ~.Chip chip:
        :rtype: ExtraMonitorSupport
        """
        return ExtraMonitorSupport(constraints=[
            ChipAndCoreConstraint(x=chip.x, y=chip.y)])

    @staticmethod
    def __new_mach_monitor(chip):
        """
        :param ~.Chip chip:
        :rtype: ExtraMonitorSupportMachineVertex
        """
        return ExtraMonitorSupportMachineVertex(
            constraints=[ChipAndCoreConstraint(x=chip.x, y=chip.y)],
            app_vertex=None)

    def __new_app_gatherer(self, ethernet_chip, vertex_to_chip_map):
        """
        :param ~.Chip ethernet_chip:
        :param dict vertex_to_chip_map:
        :rtype: DataSpeedUpPacketGather
        """
        return DataSpeedUpPacketGather(
            x=ethernet_chip.x, y=ethernet_chip.y,
            ip_address=ethernet_chip.ip_address,
            constraints=[ChipAndCoreConstraint(
                x=ethernet_chip.x, y=ethernet_chip.y)],
            extra_monitors_by_chip=vertex_to_chip_map,
            report_default_directory=self._report_dir,
            write_data_speed_up_reports=self._write_reports)

    def __new_mach_gatherer(self, ethernet_chip, vertex_to_chip_map):
        """
        :param ~.Chip ethernet_chip:
        :param dict vertex_to_chip_map:
        :rtype: DataSpeedUpPacketGatherMachineVertex
        """
        return DataSpeedUpPacketGatherMachineVertex(
            x=ethernet_chip.x, y=ethernet_chip.y,
            ip_address=ethernet_chip.ip_address,
            constraints=[ChipAndCoreConstraint(
                x=ethernet_chip.x, y=ethernet_chip.y)],
            extra_monitors_by_chip=vertex_to_chip_map,
            report_default_directory=self._report_dir,
            write_data_speed_up_reports=self._write_reports)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.graphs.application import ApplicationEdge
from pacman.model.graphs.machine import MachineEdge
from spinn_front_end_common.utilities.exceptions import ConfigurationException


class InsertEdgesToLivePacketGatherers(object):
    """ Add edges from the recorded vertices to the local Live PacketGatherers.
    """

    __slots__ = [
        # the mapping of LPG parameters to machine vertices
        "_lpg_to_vertex",
        # the SpiNNaker machine
        "_machine",
        # the placements object
        "_placements"
    ]

    def __call__(
            self, live_packet_gatherer_parameters, placements,
            live_packet_gatherers_to_vertex_mapping, machine,
            machine_graph, application_graph=None, n_keys_map=None):
        """
        :param live_packet_gatherer_parameters: the set of parameters
        :type live_packet_gatherer_parameters:
            dict(LivePacketGatherParameters,
            list(tuple(~pacman.model.graphs.AbstractVertex, list(str))))
        :param ~pacman.model.placements.Placements placements:
            the placements object
        :param live_packet_gatherers_to_vertex_mapping:
            the mapping of LPG parameters and the machine vertices associated
            with it
        :type live_packet_gatherers_to_vertex_mapping:
            dict(LivePacketGatherParameters,
            dict(tuple(int,int),LivePacketGatherMachineVertex))
        :param ~spinn_machine.Machine machine: the SpiNNaker machine
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            the machine graph
        :param application_graph: the application graph
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        :param n_keys_map: key map
        :type n_keys_map:
            ~pacman.model.routing_info.DictBasedMachinePartitionNKeysMap
        """
        # pylint: disable=too-many-arguments, attribute-defined-outside-init

        # These are all contextual, and unmodified by this algorithm
        self._lpg_to_vertex = live_packet_gatherers_to_vertex_mapping
        self._machine = machine
        self._placements = placements

        progress = ProgressBar(
            live_packet_gatherer_parameters,
            string_describing_what_being_progressed=(
                "Adding edges to the machine graph between the vertices to "
                "which live output has been requested and its local Live "
                "Packet Gatherer"))

        if application_graph is None:
            for lpg_params in progress.over(live_packet_gatherer_parameters):
                # locate vertices to connect to a LPG with these params
                for vertex, p_ids in live_packet_gatherer_parameters[
                        lpg_params]:
                    self._connect_lpg_vertex_in_machine_graph(
                        machine_graph, vertex, lpg_params, p_ids, n_keys_map)
        else:
            for lpg_params in progress.over(live_packet_gatherer_parameters):
                # locate vertices to connect to a LPG with these params
                for vertex, p_ids in live_packet_gatherer_parameters[
                        lpg_params]:
                    self._connect_lpg_vertex_in_application_graph(
                        application_graph, machine_graph, vertex, lpg_params,
                        p_ids, n_keys_map)

    @staticmethod
    def _process_partitions(partitions, n_keys_map):
        """ helper method to add to n keys map

        :param set partitions: the partitions to add keys for
        :param ~.DictBasedMachinePartitionNKeysMap n_keys_map: the key map
        :return:
        """
        for partition in partitions:
            n_keys = partition.pre_vertex.get_n_keys_for_partition(
                partition)
            n_keys_map.set_n_keys_for_partition(partition, n_keys)

    def _connect_lpg_vertex_in_application_graph(
            self, app_graph, m_graph, app_vertex, lpg_params, p_ids,
            n_keys_map):
        """
        :param ~.ApplicationGraph app_graph:
        :param ~.MachineGraph m_graph:
        :param ~.ApplicationVertex app_vertex:
        :param LivePacketGatherParameters lpg_params:
        :param list(str) p_ids:
        :param ~.DictBasedMachinePartitionNKeysMap n_keys_map:
        """
        if not app_vertex.machine_vertices or not p_ids:
            return

        # flag for ensuring we don't add the edge to the app graph many times
        app_edges = dict()
        m_partitions = set()

        # iterate through the associated machine vertices
        for vertex in app_vertex.machine_vertices:
            lpg = self._find_closest_live_packet_gatherer(vertex, lpg_params)

            # if not yet built the app edges, add them now
            # has to be postponed until we know the LPG
            if not app_edges:
                for partition_id in p_ids:
                    app_edge = ApplicationEdge(app_vertex, lpg.app_vertex)
                    app_graph.add_edge(app_edge, partition_id)
                    app_edges[partition_id] = app_edge

            # add a edge between the closest LPG and the app_vertex
            for partition_id in p_ids:
                machine_edge = MachineEdge(
                    vertex, lpg, app_edge=app_edges[partition_id])
                m_graph.add_edge(machine_edge, partition_id)

                # add to n_keys_map if needed
                if n_keys_map:
                    m_partitions.add(m_graph.get_outgoing_partition_for_edge(
                        machine_edge))
        self._process_partitions(m_partitions, n_keys_map)

    def _connect_lpg_vertex_in_machine_graph(
            self, m_graph, m_vertex, lpg_params, p_ids, n_keys_map):
        """
        :param ~.MachineGraph m_graph:
        :param ~.MachineVertex m_vertex:
        :param LivePacketGatherParameters lpg_params:
        :param list(str) p_ids:
        :param ~.DictBasedMachinePartitionNKeysMap n_keys_map:
        """
        # Find all Live Gatherer machine vertices
        lpg = self._find_closest_live_packet_gatherer(m_vertex, lpg_params)
        partitions = set()

        # add edges between the closest LPG and the vertex
        for partition_id in p_ids:
            edge = MachineEdge(m_vertex, lpg)
            m_graph.add_edge(edge, partition_id)
            # add to n_keys_map if needed
            if n_keys_map:
                partitions.add(m_graph.get_outgoing_partition_for_edge(edge))
        self._process_partitions(partitions, n_keys_map)

    def _find_closest_live_packet_gatherer(self, m_vertex, lpg_params):
        """ Locates the LPG on the nearest Ethernet-connected chip to the\
            machine vertex in question, or the LPG on 0, 0 if a closer one\
            can't be found.

        :param ~.MachineVertex m_vertex:
            the machine vertex to locate the nearest LPG to
        :param LivePacketGatherParameters lpg_params:
            parameters to decide what LPG is to be used
        :return: the local LPG
        :rtype: LivePacketGatherMachineVertex
        :raise ConfigurationException: if a local gatherer cannot be found
        """
        # locate location of vertex in machine
        placement = self._placements.get_placement_of_vertex(m_vertex)
        chip = self._machine.get_chip_at(placement.x, placement.y)

        # locate closest LPG
        machine_lpgs = self._lpg_to_vertex[lpg_params]
        chip_key = (chip.nearest_ethernet_x, chip.nearest_ethernet_y)
        if chip_key in machine_lpgs:
            return machine_lpgs[chip_key]

        # Fallback to the root (better than total failure)
        if (0, 0) in machine_lpgs:
            return machine_lpgs[0, 0]

        # if got through all LPG vertices and not found the right one. go BOOM
        raise ConfigurationException(
            "Cannot find a Live Packet Gatherer from {} for the vertex {}"
            " located on {}:{}".format(
                machine_lpgs, m_vertex, chip.x, chip.y))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from spinn_front_end_common.utilities.utility_objs import ProvenanceDataItem

#: The simple properties of PowerUsed object to be reported
_BASIC_PROPERTIES = (
    # Counts
    "num_chips", "num_cores", "num_fpgas", "num_frames",
    # Times (in seconds)
    "total_time_secs", "booted_time_secs", "mapping_time_secs",
    "data_gen_time_secs", "loading_time_secs", "exec_time_secs",
    "saving_time_secs",
    # Energies (in Joules)
    "total_energy_joules", "baseline_joules",
    "fpga_total_energy_joules", "fpga_exec_energy_joules",
    "packet_joules", "mapping_joules", "data_gen_joules",
    "loading_joules", "chip_energy_joules", "saving_joules")
#: The main provenance key we use
_PROV_KEY = "power_provenance"


class EnergyProvenanceReporter(object):
    """ Converts the power usage information into provenance data.
    """

    __slots__ = []

    def __call__(self, power_used, placements):
        """
        :param PowerUsed power_used:
            The computed basic power consumption information
        :param ~pacman.model.placements.Placements placements:
            Used for describing what a core was actually doing
        :rtype: list(ProvenanceDataItem)
        """
        prov_items = [
            ProvenanceDataItem(
                [_PROV_KEY, self.__prop_name(prop)],
                getattr(power_used, prop))
            for prop in _BASIC_PROPERTIES]
        prov_items.extend(
            ProvenanceDataItem(
                [_PROV_KEY, self.__core_name(placements, x, y, p)],
                power_used.get_core_active_energy_joules(x, y, p))
            for x, y, p in power_used.active_cores)
        prov_items.extend(
            ProvenanceDataItem(
                [_PROV_KEY, self.__router_name(x, y)],
                power_used.get_router_active_energy_joules(x, y))
            for x, y in power_used.active_routers)
        return prov_items

    @staticmethod
    def __prop_name(name):
        name = re.sub(r"_time_secs$", r" time (seconds)", name)
        return re.sub(r"(_energy)?_joules", r" energy (Joules)", name)

    @staticmethod
    def __core_name(placements, x, y, p):
        """
        :param ~.Placements placements:
        :rtype: str
        """
        if p == 0:
            # SCAMP always runs on virtual core zero, by definition
            return "SCAMP(OS)@{},{},{} energy (Joules)".format(x, y, p)
        if placements.is_processor_occupied(x, y, p):
            vtx = placements.get_vertex_on_processor(x, y, p)
            return "{} energy (Joules)".format(vtx.label)
        return "core@{},{},{} energy (Joules)".format(x, y, p)

    @staticmethod
    def __router_name(x, y):
        return "router@{},{} energy (Joules)".format(x, y)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections import defaultdict
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

__version__ = None
exec(open("spinn_front_end_common/_version.py").read())
assert __version__

# Build a list of all project modules, as well as supplementary files
main_package = "spinn_front_end_common"
extensions = {".aplx", ".boot", ".cfg", ".json", ".sql", ".template", ".xml",
              ".xsd", ".dict"}
main_package_dir = os.path.join(os.path.dirname(__file__), main_package)
start = len(main_package_dir)
packages = []
package_data = defaultdict(list)
for dirname, dirnames, filenames in os.walk(main_package_dir):
    if '__init__.py' in filenames:
        package = "{}{}".format(
            main_package, dirname[start:].replace(os.sep, '.'))
        packages.append(package)
    for filename in filenames:
        _, ext = os.path.splitext(filename)
        if ext in extensions:
            package = "{}{}".format(
                main_package, dirname[start:].replace(os.sep, '.'))
            package_data[package].append(filename)

setup(
    name="SpiNNFrontEndCommon",
    version=__version__,
    description="Common SpiNNaker Front end functions",
    url="https://github.com/SpiNNakerManchester/SpiNNFrontEndCommon",
    classifiers=[
        "Development Status :: 5 - Production/Stable",

        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",

        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",

        "Natural Language :: English",

        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",

        "Programming Language :: C",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=packages,
    package_data=package_data,
    install_requires=['SpiNNUtilities == 1!6.0.0',
                      'SpiNNMachine == 1!6.0.0',
                      'SpiNNMan == 1!6.0.0',
                      'SpiNNaker_PACMAN == 1!6.0.0',
                      'SpiNNaker_DataSpecification == 1!6.0.0',
                      'spalloc == 1!6.0.0',
                      'requests >= 2.4.1',
                      "numpy > 1.13, < 1.20; python_version == '3.6'",
                      "numpy > 1.13, < 1.21; python_version == '3.7'",
                      "numpy; python_version >= '3.8'",
                      "scipy >= 0.16.0, < 1.6; python_version == '3.6'",
                      "scipy >= 0.16.0; python_version >= '3.7'"],
    extras_require={
        'plotting': [
            'matplotlib',
            'seaborn',
        ]},
    entry_points={
        "console_scripts": [
            "spinnaker_router_provenance_mapper="
            "spinn_front_end_common.interface.provenance.router_prov_mapper:"
            "main [plotting]",
        ]},
    maintainer="SpiNNakerTeam",
    maintainer_email="spinnakerusers@googlegroups.com"
)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import struct
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from data_specification.constants import MAX_MEM_REGIONS
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

logger = FormatAdapter(logging.getLogger(__name__))
_ONE_WORD = struct.Struct("<I")
MEM_MAP_SUBDIR_NAME = "memory_map_reports"
MEM_MAP_FILENAME = "memory_map_from_processor_{0:d}_{1:d}_{2:d}.txt"
REGION_HEADER_SIZE = 2 * BYTES_PER_WORD


class MemoryMapOnHostChipReport(object):
    """ Report on memory usage. Creates a report that states where in SDRAM \
        each region is (read from machine)
    """

    def __call__(self, report_default_directory, dsg_targets, transceiver):
        """
        :param str report_default_directory:
            the folder where reports are written
        :param dict(tuple(int,int,int),...) dsg_targets:
            the map between placement and file writer
        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan instance
        """
        directory_name = os.path.join(
            report_default_directory, MEM_MAP_SUBDIR_NAME)
        if not os.path.exists(directory_name):
            os.makedirs(directory_name)

        progress = ProgressBar(dsg_targets, "Writing memory map reports")
        for (x, y, p) in progress.over(dsg_targets):
            file_name = os.path.join(
                directory_name, MEM_MAP_FILENAME.format(x, y, p))
            try:
                with open(file_name, "w") as f:
                    self._describe_mem_map(f, transceiver, x, y, p)
            except IOError:
                logger.exception("Generate_placement_reports: Can't open file"
                                 " {} for writing.", file_name)

    def _describe_mem_map(self, f, txrx, x, y, p):
        """
        :param ~spinnman.transceiver.Transceiver txrx:
        """
        # pylint: disable=too-many-arguments
        # Read the memory map data from the given core
        region_table_addr = self._get_region_table_addr(txrx, x, y, p)
        memmap_data = txrx.read_memory(
            x, y, region_table_addr, BYTES_PER_WORD * MAX_MEM_REGIONS)

        # Convert the map to a human-readable description
        f.write("On chip data specification executor\n\n")
        for i in range(MAX_MEM_REGIONS):
            region_address, = _ONE_WORD.unpack_from(
                memmap_data, i * BYTES_PER_WORD)
            f.write("Region {0:d}:\n\t start address: 0x{1:x}\n\n".format(
                i, region_address))

    def _get_region_table_addr(self, txrx, x, y, p):
        """
        :param ~spinnman.transceiver.Transceiver txrx:
        """
        user_0_addr = txrx.get_user_0_register_address_from_core(p)
        return txrx.read_word(x, y, user_0_addr) + REGION_HEADER_SIZE



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pacman.model.graphs.application import ApplicationEdge, ApplicationVertex
from pacman.model.graphs.application.abstract import (
    AbstractOneAppOneMachineVertex)
from .command_sender_machine_vertex import CommandSenderMachineVertex


class CommandSender(
        AbstractOneAppOneMachineVertex):
    """ A utility for sending commands to a vertex (possibly an external\
        device) at fixed times in the simulation or in response to \
        simulation events (e.g., starting and stopping).
    """

    def __init__(self, label, constraints):
        """
        :param str label: The label of this vertex
        :param constraints: Any initial constraints to this vertex
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint)
        """
        super().__init__(
            CommandSenderMachineVertex(label, constraints, self),
            label, constraints)

    def add_commands(
            self, start_resume_commands, pause_stop_commands,
            timed_commands, vertex_to_send_to):
        """ Add commands to be sent down a given edge

        :param iterable(MultiCastCommand) start_resume_commands:
            The commands to send when the simulation starts or resumes from
            pause
        :param iterable(MultiCastCommand) pause_stop_commands:
            The commands to send when the simulation stops or pauses after
            running
        :param iterable(MultiCastCommand) timed_commands:
            The commands to send at specific times
        :param ~pacman.model.graphs.AbstractVertex vertex_to_send_to:
            The vertex these commands are to be sent to
        """
        self._machine_vertex.add_commands(
            start_resume_commands, pause_stop_commands, timed_commands,
            vertex_to_send_to)

    def edges_and_partitions(self):
        """ Construct application edges from this vertex to the app vertices\
            that this vertex knows how to target (and has keys allocated for).

        :return: edges, partition IDs
        :rtype: tuple(list(~pacman.model.graphs.application.ApplicationEdge),
            list(str))
        """
        return self._machine_vertex._get_edges_and_partitions(
            self, ApplicationVertex, ApplicationEdge)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import (
    AbstractBase, abstractproperty, abstractmethod)


class SimulatorInterface(object, metaclass=AbstractBase):

    __slots__ = ()

    @abstractmethod
    def add_socket_address(self, socket_address):
        """ Add the address of a socket used in the run notification protocol.

        :param ~spinn_utilities.socket_address.SocketAddress socket_address:
            The address of the socket
        :rtype: None
        """

    @abstractproperty
    def buffer_manager(self):
        """ The buffer manager being used for loading/extracting buffers

        :rtype:
            ~spinn_front_end_common.interface.buffer_management.BufferManager
        """

    @abstractproperty
    def config(self):
        """ Provides access to the configuration for front end interfaces.

        :rtype: ~spinn_front_end_common.interface.config_handler.ConfigHandler
        """

    @abstractproperty
    def has_ran(self):
        """ Whether the simulation has executed anything at all.

        :rtype: bool
        """

    @abstractmethod
    def verify_not_running(self):
        """ Verify that the simulator is in a state where it can start running.
        """

    @abstractproperty
    def machine(self):
        """ The python machine description object.

        :rtype: ~spinn_machine.Machine
        """

    @abstractproperty
    def machine_time_step(self):
        """ The machine timestep, in microseconds.

        :rtype: int
        """

    @abstractproperty
    def no_machine_time_steps(self):
        """ The number of machine time steps.

        :rtype: int
        """

    @abstractproperty
    def placements(self):
        """ Where machine vertices are placed on the machine.

        :rtype: ~pacman.model.placements.Placements
        """

    @abstractproperty
    def tags(self):
        """
        :rtype: ~pacman.model.tags.Tags
        """

    @abstractproperty
    def time_scale_factor(self):
        """
        :rtype: int
        """

    @abstractmethod
    def run(self, run_time, sync_time=0.0):
        """ Run a simulation for a fixed amount of time

        :param int run_time: the run duration in milliseconds.
        :param float sync_time:
            If not 0, this specifies that the simulation should pause after
            this duration.  The continue_simulation() method must then be
            called for the simulation to continue.
        """

    @abstractmethod
    def stop(self):
        """ End running of the simulation.
        """

    @abstractmethod
    def stop_run(self):
        """ End the running of a simulation that has been started with
            run_forever
        """

    @abstractmethod
    def continue_simulation(self):
        """ Continue a simulation that has been started in stepped mode
        """

    @abstractproperty
    def transceiver(self):
        """ How to talk to the machine.

        :rtype: ~spinnman.transceiver.Transceiver
        """

    @abstractproperty
    def use_virtual_board(self):
        """
        :rtype: bool
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from spinnman.model import ExecutableTargets
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from pacman.model.graphs import AbstractVirtual
from spinn_front_end_common.utilities.exceptions import (
    ExecutableNotFoundException)
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary

logger = FormatAdapter(logging.getLogger(__name__))


class GraphBinaryGatherer(object):
    """ Extracts binaries to be executed.
    """

    __slots__ = ["_exe_finder", "_exe_targets"]

    def __init__(self):
        self._exe_finder = None
        self._exe_targets = None

    def __call__(self, placements, graph, executable_finder):
        """
        :param ~pacman.model.placements.Placements placements:
        :param ~pacman.model.graphs.machine.MachineGraph graph:
        :param executable_finder:
        :type executable_finder:
            ~spinn_utilities.executable_finder.ExecutableFinder
        :rtype: ExecutableTargets
        """
        self._exe_finder = executable_finder
        self._exe_targets = ExecutableTargets()
        progress = ProgressBar(graph.n_vertices, "Finding binaries")
        for vertex in progress.over(graph.vertices):
            placement = placements.get_placement_of_vertex(vertex)
            self.__get_binary(placement, vertex)

        return self._exe_targets

    def __get_binary(self, placement, vertex):
        """
        :param ~.Placement placement:
        :param ~.AbstractVertex vertex:
        """
        # if the vertex cannot be executed, ignore it
        if not isinstance(vertex, AbstractHasAssociatedBinary):
            if not isinstance(vertex, AbstractVirtual):
                msg = (
                    "Vertex {} does not implement either "
                    "AbstractHasAssociatedBinary or AbstractVirtual. So it is "
                    "unclear if it should or should not have a binary")
                logger.error(msg.format(vertex), vertex)
            return

        # Get name of binary from vertex
        binary_name = vertex.get_binary_file_name()
        exec_type = vertex.get_binary_start_type()

        # Attempt to find this within search paths
        binary_path = self._exe_finder.get_executable_path(binary_name)
        if binary_path is None:
            raise ExecutableNotFoundException(binary_name)

        self._exe_targets.add_processor(
            binary_path, placement.x, placement.y, placement.p, exec_type)



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class BitFieldSummary(object):
    """ Summary description of generated bitfields.
    """

    def __init__(
            self, total_merged, max_per_chip, lowest_per_chip, total_to_merge,
            max_to_merge_per_chip, low_to_merge_per_chip,
            average_per_chip_merged, average_per_chip_to_merge):
        """
        :param int total_merged:
        :param int max_per_chip:
        :param int lowest_per_chip:
        :param int total_to_merge:
        :param int max_to_merge_per_chip:
        :param int low_to_merge_per_chip:
        :param float average_per_chip_merged:
        :param float average_per_chip_to_merge:
        """
        self._total_merged = total_merged
        self._max_per_chip = max_per_chip
        self._lowest_per_chip = lowest_per_chip
        self._total_to_merge = total_to_merge
        self._max_to_merge_per_chip = max_to_merge_per_chip
        self._low_to_merge_per_chip = low_to_merge_per_chip
        self._average_per_chip_merged = average_per_chip_merged
        self._average_per_chip_to_merge = average_per_chip_to_merge

    @property
    def total_merged(self):
        """
        :rtype: int
        """
        return self._total_merged

    @property
    def max_per_chip(self):
        """
        :rtype: int
        """
        return self._max_per_chip

    @property
    def lowest_per_chip(self):
        """
        :rtype: int
        """
        return self._lowest_per_chip

    @property
    def total_to_merge(self):
        """
        :rtype: int
        """
        return self._total_to_merge

    @property
    def max_to_merge_per_chip(self):
        """
        :rtype: int
        """
        return self._max_to_merge_per_chip

    @property
    def low_to_merge_per_chip(self):
        """
        :rtype: int
        """
        return self._low_to_merge_per_chip

    @property
    def average_per_chip_merged(self):
        """
        :rtype: float
        """
        return self._average_per_chip_merged

    @property
    def average_per_chip_to_merge(self):
        """
        :rtype: float
        """
        return self._average_per_chip_to_merge



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sqlite3
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB
from spinn_front_end_common.utilities.utility_objs import DataWritten

DB_NAME = "ds.sqlite3"
_DDL_FILE = os.path.join(os.path.dirname(__file__), "dse.sql")
logger = FormatAdapter(logging.getLogger(__name__))


class DsSqlliteDatabase(SQLiteDB):
    __slots__ = [
        # The machine cached for getting the "ethernet"s
        "_machine",
        # The root ethernet id if required
        "_root_ethernet_id"
    ]

    def __init__(self, machine, report_folder, init=None):
        """
        :param ~spinn_machine.Machine machine:
        :param str report_folder:
        :param init:
        :type init: bool or None
        """
        self._machine = machine
        database_file = os.path.join(report_folder, DB_NAME)

        if init is None:
            init = not os.path.exists(database_file)

        super().__init__(database_file, ddl_file=_DDL_FILE if init else None)
        if init:
            self.__init_db_contents()
        self._root_ethernet_id = self.__find_root_id()

    def __init_db_contents(self):
        """ Set up the database contents from the machine. """
        with self.transaction() as cursor:
            cursor.executemany(
                """
                INSERT INTO ethernet(
                    ethernet_x, ethernet_y, ip_address)
                VALUES(?, ?, ?)
                """, (
                    (ethernet.x, ethernet.y, ethernet.ip_address)
                    for ethernet in self._machine.ethernet_connected_chips))

    def __find_root_id(self):
        first_x = first_y = root_id = None
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT ethernet_id, ethernet_x, ethernet_y FROM ethernet
                    ORDER BY ethernet_x, ethernet_y
                    LIMIT 1
                    """):
                root_id = row["ethernet_id"]
                first_x = row["ethernet_x"]
                first_y = row["ethernet_y"]
        if root_id is None:
            # Should only be reachable for an empty machine
            raise Exception("No ethernet chip found")
        if first_x or first_y:
            logger.warning(
                "No Ethernet chip found at 0,0 using {},{} "
                "for all boards with no IP address.", first_x, first_y)
        return root_id

    def clear_ds(self):
        """ Clear all saved data specification data
        """
        with self.transaction() as cursor:
            cursor.execute(
                """
                DELETE FROM core
                """)

    def save_ds(self, core_x, core_y, core_p, ds):
        """
        :param int core_x: x of the core ds applies to
        :param int core_y: y of the core ds applies to
        :param int p: p of the core ds applies to
        :param bytearray ds: the data spec as byte code
        """
        chip = self._machine.get_chip_at(core_x, core_y)
        with self.transaction() as cursor:
            cursor.execute(
                """
                INSERT INTO core(
                    x, y, processor, content, ethernet_id)
                VALUES(?, ?, ?, ?, (
                    SELECT COALESCE((
                        SELECT ethernet_id FROM ethernet
                        WHERE ethernet_x = ? AND ethernet_y = ?
                    ), ?)))
                """, (
                    core_x, core_y, core_p, sqlite3.Binary(ds),
                    chip.nearest_ethernet_x, chip.nearest_ethernet_y,
                    self._root_ethernet_id))

    def get_ds(self, x, y, p):
        """ Retrieves the data spec as byte code for this core.

        :param int x: core x
        :param int y: core y
        :param int p: core p
        :return: data spec as byte code
        :rtype: bytearray
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT content FROM core
                    WHERE x = ? AND y = ? AND processor = ?
                    LIMIT 1
                    """, (x, y, p)):
                return row["content"]
        return b""

    def ds_iteritems(self):
        """ Yields the keys and values for the DS data

        .. note:
            Do not use the database for anything else while iterating.

        :return: Yields the (x, y, p) and saved ds pairs
        :rtype: iterable(tuple(tuple(int, int, int), bytearray))
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT x, y, processor, content FROM core
                    WHERE content IS NOT NULL
                    """):
                yield (row["x"], row["y"], row["processor"]), row["content"]

    def ds_n_cores(self):
        """ Returns the number for cores there is a ds saved for

        :rtype: int
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT COUNT(*) as count FROM core
                    WHERE content IS NOT NULL
                    LIMIT 1
                    """):
                return row["count"]
        raise Exception("Count query failed")

    def ds_set_app_id(self, app_id):
        """ Sets the same app_id for all rows that have ds content

        :param int app_id: value to set
        """
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE core SET
                    app_id = ?
                WHERE content IS NOT NULL
                """, (app_id,))

    def ds_get_app_id(self, x, y, p):
        """ Gets the app_id set for this core

        :param int x: core x
        :param int y: core y
        :param int p: core
        :rtype: int
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT app_id FROM core
                    WHERE x = ? AND y = ? AND processor = ?
                    LIMIT 1
                    """, (x, y, p)):
                return row["app_id"]
        return None

    def ds_mark_as_system(self, core_list):
        """ Flags a list of processors as running system binaries.

        :param iterable(tuple(int,int,int)) core_list:
            list of (core x, core y, core p)
        """
        with self.transaction() as cursor:
            cursor.executemany(
                """
                UPDATE core SET
                    is_system = 1
                WHERE x = ? AND y = ? AND processor = ?
                """, core_list)

    def _row_to_info(self, row):
        """
        :param ~sqlite3.Row row:
        :rtype: DataWritten
        """
        return DataWritten(start_address=row["start_address"],
                           memory_used=row["memory_used"],
                           memory_written=row["memory_written"])

    def get_write_info(self, x, y, p):
        """ Gets the provenance returned by the Data Spec executor.

        :param int x: core x
        :param int y: core y
        :param int p: core p
        :rtype: DataWritten
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT start_address, memory_used, memory_written
                    FROM core
                    WHERE x = ? AND y = ? AND processor = ?
                    LIMIT 1
                    """, (x, y, p)):
                return self._row_to_info(row)
        raise ValueError(f"No info for {x}:{y}:{p}")

    def set_write_info(self, x, y, p, info):
        """ Sets the provenance returned by the Data Spec executor.

        :param int x: core x
        :param int y: core y
        :param int p: core p
        :param DataWritten info:
        """
        if isinstance(info, DataWritten):
            start = info.start_address
            used = info.memory_used
            written = info.memory_written
        else:
            start = info["start_address"]
            used = info["memory_used"]
            written = info["memory_written"]
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE core SET
                    start_address = ?,
                    memory_used = ?,
                    memory_written = ?
                WHERE x = ? AND y = ? AND processor = ?
                """, (start, used, written, x, y, p))
            if cursor.rowcount == 0:
                chip = self._machine.get_chip_at(x, y)
                cursor.execute(
                    """
                    INSERT INTO core(
                        x, y, processor, start_address,
                        memory_used, memory_written, ethernet_id)
                    VALUES(?, ?, ?, ?, ?, ?, (
                        SELECT COALESCE((
                            SELECT ethernet_id FROM ethernet
                            WHERE ethernet_x = ? AND ethernet_y = ?
                        ), ?)))
                    """, (
                        x, y, p, start, used, written, chip.nearest_ethernet_x,
                        chip.nearest_ethernet_y, self._root_ethernet_id))

    def set_size_info(self, x, y, p, memory_used):
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE core SET
                    memory_used = ?
                WHERE x = ? AND y = ? AND processor = ?
                """, (memory_used, x, y, p))
            if cursor.rowcount == 0:
                chip = self._machine.get_chip_at(x, y)
                cursor.execute(
                    """
                    INSERT INTO core(
                        x, y, processor, memory_used, ethernet_id)
                    VALUES(?, ?, ?, ?, (
                        SELECT COALESCE((
                            SELECT ethernet_id FROM ethernet
                            WHERE ethernet_x = ? AND ethernet_y = ?
                        ), ?)))
                    """, (
                        x, y, p, int(memory_used),
                        chip.nearest_ethernet_x, chip.nearest_ethernet_y,
                        self._root_ethernet_id))

    def clear_write_info(self):
        """ Clears the provenance for all rows.
        """
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE core SET
                    start_address = NULL,
                    memory_used = NULL,
                    memory_written = NULL
                """)

    def info_n_cores(self):
        """ Returns the number for cores there is a info saved for.

        :rtype: int
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT count(*) as count FROM core
                    WHERE start_address IS NOT NULL
                    LIMIT 1
                    """):
                return row["count"]
        raise Exception("Count query failed")

    def info_iteritems(self):
        """ Yields the keys and values for the Info data.

        .. note:
            A DB transaction may be held while this iterator is processing.
            Reentrant use of this class is not supported.

        :return: Yields the (x, y, p) and DataWritten
        :rtype: iterable(tuple(tuple(int, int, int), DataWritten))
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT
                        x, y, processor,
                        start_address, memory_used, memory_written
                    FROM core
                    WHERE start_address IS NOT NULL
                    """):
                yield (row["x"], row["y"], row["processor"]), \
                      self._row_to_info(row)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict

from data_specification.constants import APP_PTR_TABLE_BYTE_SIZE
from spinn_utilities.progress_bar import ProgressBar
from data_specification import DataSpecificationGenerator
from data_specification.utility_calls import get_report_writer
from spinn_front_end_common.abstract_models import (
    AbstractRewritesDataSpecification, AbstractGeneratesDataSpecification)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.interface.ds.data_specification_targets import (
    DataSpecificationTargets)


class GraphDataSpecificationWriter(object):
    """ Executes the data specification generation step.
    """

    __slots__ = (
        # Dict of SDRAM usage by chip coordinates
        "_sdram_usage",
        # Dict of list of region sizes by core coordinates
        "_region_sizes",
        # Dict of list of vertices by chip coordinates
        "_vertices_by_chip",
        # spinnmachine instance
        "_machine",
        # hostname
        "_hostname",
        # directory where reports go
        "_report_dir",
        # bool for writing texts
        "_write_text",
    )

    def __init__(self):
        self._sdram_usage = defaultdict(lambda: 0)
        self._region_sizes = dict()
        self._vertices_by_chip = defaultdict(list)

    def __call__(
            self, placements, hostname,
            report_default_directory, write_text_specs,
            machine, data_n_timesteps, placement_order=None):
        """
        :param ~pacman.model.placements.Placements placements:
            placements of machine graph to cores
        :param str hostname: SpiNNaker machine name
        :param str report_default_directory:
            the location where reports are stored
        :param bool write_text_specs:
            True if the textual version of the specification is to be written
        :param ~spinn_machine.Machine machine:
            the python representation of the SpiNNaker machine
        :param int data_n_timesteps:
            The number of timesteps for which data space will been reserved
        :param list(~pacman.model.placements.Placement) placement_order:
            the optional order in which placements should be examined
        :return: DSG targets (map of placement tuple and filename)
        :rtype: tuple(DataSpecificationTargets, dict(tuple(int,int,int), int))
        :raises ConfigurationException:
            If the DSG asks to use more SDRAM than is available.
        """
        # pylint: disable=too-many-arguments, too-many-locals
        # pylint: disable=attribute-defined-outside-init
        self._machine = machine
        self._hostname = hostname
        self._report_dir = report_default_directory
        self._write_text = write_text_specs

        # iterate though vertices and call generate_data_spec for each
        # vertex
        targets = DataSpecificationTargets(machine, self._report_dir)

        if placement_order is None:
            placement_order = placements.placements

        progress = ProgressBar(
            placements.n_placements, "Generating data specifications")
        vertices_to_reset = list()
        for placement in progress.over(placement_order):
            # Try to generate the data spec for the placement
            vertex = placement.vertex
            generated = self.__generate_data_spec_for_vertices(
                placement, vertex, targets, data_n_timesteps)

            if generated and isinstance(
                    vertex, AbstractRewritesDataSpecification):
                vertices_to_reset.append(vertex)

            # If the spec wasn't generated directly, and there is an
            # application vertex, try with that
            if not generated and vertex.app_vertex is not None:
                generated = self.__generate_data_spec_for_vertices(
                    placement, vertex.app_vertex, targets, data_n_timesteps)
                if generated and isinstance(
                        vertex.app_vertex, AbstractRewritesDataSpecification):
                    vertices_to_reset.append(vertex.app_vertex)

        # Ensure that the vertices know their regions have been reloaded
        for vertex in vertices_to_reset:
            vertex.set_reload_required(False)

        return targets, self._region_sizes

    def __generate_data_spec_for_vertices(
            self, pl, vertex, targets, data_n_timesteps):
        """
        :param ~.Placement pl: placement of machine graph to cores
        :param ~.AbstractVertex vertex: the specific vertex to write DSG for.
        :param DataSpecificationTargets targets:
        :return: True if the vertex was data spec-able, False otherwise
        :rtype: bool
        :raises ConfigurationException: if things don't fit
        """
        # if the vertex can generate a DSG, call it
        if not isinstance(vertex, AbstractGeneratesDataSpecification):
            return False

        with targets.create_data_spec(pl.x, pl.y, pl.p) as data_writer:
            report_writer = get_report_writer(
                pl.x, pl.y, pl.p, self._hostname,
                self._report_dir, self._write_text)
            spec = DataSpecificationGenerator(data_writer, report_writer)

            # generate the DSG file
            vertex.generate_data_specification(spec, pl)

            # Check the memory usage
            self._region_sizes[pl.x, pl.y, pl.p] = (
                APP_PTR_TABLE_BYTE_SIZE + sum(spec.region_sizes))

            # extracts the int from the numpy data type generated
            if not isinstance(self._region_sizes[pl.x, pl.y, pl.p], int):
                self._region_sizes[pl.x, pl.y, pl.p] =\
                    self._region_sizes[pl.x, pl.y, pl.p].item()

            self._vertices_by_chip[pl.x, pl.y].append(pl.vertex)
            self._sdram_usage[pl.x, pl.y] += sum(spec.region_sizes)
            if (self._sdram_usage[pl.x, pl.y] <=
                    self._machine.get_chip_at(pl.x, pl.y).sdram.size):
                return True

        # creating the error message which contains the memory usage of
        #  what each core within the chip uses and its original
        # estimate.
        memory_usage = "\n".join((
            "    {}: {} (total={}, estimated={})".format(
                vert, self._region_sizes[pl.x, pl.y, pl.p],
                sum(self._region_sizes[pl.x, pl.y, pl.p]),
                vert.resources_required.sdram.get_total_sdram(
                    data_n_timesteps))
            for vert in self._vertices_by_chip[pl.x, pl.y]))

        raise ConfigurationException(
            "Too much SDRAM has been used on {}, {}.  Vertices and"
            " their usage on that chip is as follows:\n{}".format(
                pl.x, pl.y, memory_usage))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from spinn_utilities.progress_bar import ProgressBar
from pacman.utilities.utility_objs import ResourceTracker
from pacman.utilities.algorithm_utilities.placer_algorithm_utilities import (
    sort_vertices_by_known_constraints)


class GraphMeasurer(object):
    """ Works out how many chips a machine graph needs.
    """

    __slots__ = []

    def __call__(self, machine_graph, machine, plan_n_timesteps):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            The machine_graph to measure.
        :param ~spinn_machine.Machine machine:
            The machine with respect to which to partition the application
            graph.
        :param int plan_n_timesteps: Number of timesteps to plan for.
        :return: The size of the graph in number of chips.
        :rtype: int
        """

        # check that the algorithm can handle the constraints
        ResourceTracker.check_constraints(machine_graph.vertices)

        ordered_vertices = sort_vertices_by_known_constraints(
            machine_graph.vertices)

        # Iterate over vertices and allocate
        progress = ProgressBar(machine_graph.n_vertices, "Measuring the graph")

        resource_tracker = ResourceTracker(machine, plan_n_timesteps)
        for vertex in progress.over(ordered_vertices):
            resource_tracker.allocate_constrained_resources(
                vertex.resources_required, vertex.constraints)
        return len(resource_tracker.keys)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.application import ApplicationVertex


@require_subclass(ApplicationVertex)
class AbstractVertexWithEdgeToDependentVertices(
        object, metaclass=AbstractBase):
    """ A vertex with a dependent vertices, which should be connected to this\
        vertex by an edge directly to each of them
    """

    __slots__ = ()

    @abstractmethod
    def dependent_vertices(self):
        """ Return the vertices which this vertex depends upon

        :rtype: iterable(~pacman.model.graphs.application.ApplicationVertex)
        """

    @abstractmethod
    def edge_partition_identifiers_for_dependent_vertex(self, vertex):
        """ Return the dependent edge identifiers for a particular dependent\
            vertex.

        :param ~pacman.model.graphs.application.ApplicationVertex vertex:
        :rtype: iterable(str)
        """



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractSupportsBitFieldGeneration(object, metaclass=AbstractBase):
    """ Marks a vertex that can provide information about bitfields it wants \
        generated on-chip.
    """
    __slots__ = ()

    @abstractmethod
    def bit_field_base_address(self, transceiver, placement):
        """ Returns the SDRAM address for the bit field table data.

        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :return: the SDRAM address for the bitfield address
        :rtype: int
        """

    @abstractmethod
    def bit_field_builder_region(self, transceiver, placement):
        """ returns the SDRAM address for the bit field builder data

        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :return: the SDRAM address for the bitfield builder data
        :rtype: int
        """
